/* x10bot
 * Copyright 2001-2002 Daniel P. Stasinski <mooooooo@scriptkitties.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <x10_api.h>
#include <tcl.h>
#include <tclDecls.h>
#include "x10bot.h"

extern int running;
extern int sockets[MAXSOCKETS];
extern Tcl_Interp *interp;
extern struct x10_status_t *x10_status;
extern int accept_socket;
extern int sock;
extern x10module modules[MAXMODULES];
extern event events[MAXEVENTS];
extern int port;
extern char *logfile;
extern int foreground;
extern time_t uptime;
extern double latitude;
extern double longitude;
extern COMMANDLIST *addedcommands;
extern int thissocket;

void do_turn(int s, int mode)
{
    char reply[BUFLEN];
    char house, *tmp_char;
    char *module;
    int unit, module_id, command;
    int value = 0;
    int isgroup;

    if (!(module = strtok(NULL, " "))) {
        sprintf(reply, "ERROR\n");
        write(s, reply, strlen(reply));
        return;
    }

    if (mode == TOGGLE) {
        if (!(tmp_char = strtok(NULL, " "))) {
            sprintf(reply, "ERROR\n");
            write(s, reply, strlen(reply));
            return;
        }

        if (strcasecmp(tmp_char, "on") == 0)
            command = X10_CMD_ON;
        else if (strcasecmp(tmp_char, "off") == 0)
            command = X10_CMD_OFF;
        else {
            sprintf(reply, "ERROR (%s)\n", tmp_char);
            write(s, reply, strlen(reply));
            return;
        }
    } else {
        if (mode == DIM)
            command = X10_CMD_DIM;
        else if (mode == BRIGHT)
            command = X10_CMD_BRIGHT;
        else {
            sprintf(reply, "ERROR (%s)\n", tmp_char);
            write(s, reply, strlen(reply));
            return;
        }
        if ((tmp_char = strtok(NULL, " ")))
            value = atoi(tmp_char);
    }

    module_id = getmodule(module);
    if (module_id == -1) {
        isgroup = FALSE;
        for (module_id = 0; module_id < MAXMODULES; module_id++) {
            if (strcasecmp(module, modules[module_id].group) == 0) {
                house = ((module_id / 16) + 'A');
                unit = (module_id % 16) + 1;
                if (x10_send_cmd(house, unit, command, value, x10_status)) {
                    sprintf(reply, "ERROR\n");
                    write(s, reply, strlen(reply));
                } else {
                    if (modules[module_id].alias != NULL)
                        sprintf(reply, "SENT %s %s %d\n", modules[module_id].alias, cmd2string(command), value);
                    else
                        sprintf(reply, "SENT %c%d %s %d\n", house, unit, cmd2string(command), value);
                    send_to_all(reply);
                }
                isgroup = TRUE;
            }
        }
        if (!isgroup) {
            sprintf(reply, "ERROR\n");
            write(s, reply, strlen(reply));
        }
        return;
    } else {

        house = ((module_id / 16) + 'A');
        unit = (module_id % 16) + 1;

        if (x10_send_cmd(house, unit, command, value, x10_status)) {
            sprintf(reply, "ERROR\n");
            write(s, reply, strlen(reply));
            return;
        }

        if (modules[module_id].alias != NULL)
            sprintf(reply, "SENT %s %s %d\n", modules[module_id].alias, cmd2string(command), value);
        else
            sprintf(reply, "SENT %c%d %s %d\n", house, unit, cmd2string(command), value);

        send_to_all(reply);

        if (command == X10_CMD_ON)
            camera_group_off(module_id, x10_status);

    }
}

void showtimers(int s)
{

    char reply[BUFLEN];
    int x;

    for (x = 0; x < MAXEVENTS; x++) {
        if (events[x].mode == SCHED_NEVER)
            continue;
        if (events[x].script == NULL)
            continue;
        sprintf(reply, "TIMER %d %d (%s) %s\n", events[x].handle, (int) events[x].timestamp, asciitime(&events[x].timestamp), events[x].script);
        write(s, reply, strlen(reply));
    }

}

void do_status_all(int s)
{

    char reply[BUFLEN];
    struct x10_object_t module;
    int i, j, module_id = 0;

    for (i = 0; i < 16; i++) {
        for (j = 0; j < 16; j++) {
            module = x10_status->objects[i][j];
            if (modules[module_id].alias != NULL)
                sprintf(reply, "STATUS %s %s %d (%s)\n", modules[module_id].alias, status2string(module.status), (int) module.timestamp, asciitime(&module.timestamp));
            else
                sprintf(reply, "STATUS %c%d %s %d (%s)\n", 'A' + i, j + 1, status2string(module.status), (int) module.timestamp, asciitime(&module.timestamp));
            write(s, reply, strlen(reply));
            module_id++;
        }
    }

}

void do_status_house(int s)
{

    char reply[BUFLEN];
    struct x10_object_t module;
    int i, j, module_id = 0;
    char *house;

    if (!(house = strtok(NULL, " "))) {
        sprintf(reply, "ERROR\n");
        write(s, reply, strlen(reply));
        return;
    }

    house[0] = toupper(house[0]);

    if (house[0] < 'A' || house[0] > 'P') {
        sprintf(reply, "ERROR\n");
        write(s, reply, strlen(reply));
        return;
    }

    i = house[0] - 'A';

    module_id = i * 16;
    for (j = 0; j < 16; j++) {
        module = x10_status->objects[i][j];
        if (modules[module_id].alias != NULL)
            sprintf(reply, "STATUS %s %s %d (%s)\n", modules[module_id].alias, status2string(module.status), (int) module.timestamp, asciitime(&module.timestamp));
        else
            sprintf(reply, "STATUS %c%d %s %d (%s)\n", 'A' + i, j + 1, status2string(module.status), (int) module.timestamp, asciitime(&module.timestamp));
        write(s, reply, strlen(reply));
        module_id++;
    }

}

void do_status(int s)
{

    char reply[BUFLEN];
    char house, *tmp_char;
    int unit, module_id;
    struct x10_object_t module;

    if (!(tmp_char = strtok(NULL, " "))) {
        sprintf(reply, "ERROR\n");
        write(s, reply, strlen(reply));
        return;
    }

    module_id = getmodule(tmp_char);
    if (module_id == -1) {
        sprintf(reply, "ERROR\n");
        write(s, reply, strlen(reply));
        return;
    }

    house = ((module_id / 16) + 'A');
    unit = (module_id % 16) + 1;

    module = x10_status->objects[house - 'A'][unit - 1];

    if (modules[module_id].alias != NULL)
        sprintf(reply, "STATUS %s %s %d (%s)\n", modules[module_id].alias, status2string(module.status), (int) module.timestamp, asciitime(&module.timestamp));
    else
        sprintf(reply, "STATUS %c%d %s %d (%s)\n", house, unit, status2string(module.status), (int) module.timestamp, asciitime(&module.timestamp));

    write(s, reply, strlen(reply));

}

void do_send(int s)
{

    char reply[BUFLEN];
    char house, *tmp_char;
    int module_id, unit, command, arg;

    if (!(tmp_char = strtok(NULL, " "))) {
        sprintf(reply, "ERROR\n");
        write(s, reply, strlen(reply));
        return;
    }

    module_id = getmodule(tmp_char);
    if (module_id == -1) {
        sprintf(reply, "ERROR\n");
        write(s, reply, strlen(reply));
        return;
    }

    house = ((module_id / 16) + 'A');
    unit = (module_id % 16) + 1;

    if (!(tmp_char = strtok(NULL, " "))) {
        sprintf(reply, "ERROR\n");
        write(s, reply, strlen(reply));
        return;
    }

    command = string2cmd(tmp_char);

    if (command == -1) {
        sprintf(reply, "ERROR\n");
        write(s, reply, strlen(reply));
        return;
    }

    if (!(tmp_char = strtok(NULL, " "))) {
        sprintf(reply, "ERROR\n");
        write(s, reply, strlen(reply));
        return;
    }

    arg = atoi(tmp_char);

    if (x10_send_cmd(house, unit, command, arg, x10_status)) {
        sprintf(reply, "ERROR\n");
        write(s, reply, strlen(reply));
        return;
    }

    if (modules[module_id].alias != NULL)
        sprintf(reply, "SENT %s %s %d\n", modules[module_id].alias, cmd2string(command), arg);
    else
        sprintf(reply, "SENT %c%d %s %d\n", house, unit, cmd2string(command), arg);

    send_to_all(reply);

}

int do_accept(int acc_sock)
{
    int sinlen, ret;
    struct sockaddr sin;

    memset((char *) &sin, 0, sizeof(sin));

    sinlen = sizeof(sin);

    ret = accept(acc_sock, &sin, &sinlen);

    if (ret != -1)
        return (ret);
    else
        return (0);

}

void send_to_all(char *string)
{
    int x;
    FILE *fp;
    time_t timep;

    timep = time(NULL);

    if (logfile != NULL) {
        fp = fopen(logfile, "a");
        fprintf(fp, "%s %s", asciitime(&timep), string);
        fclose(fp);
    }

    for (x = 0; x < MAXSOCKETS; x++) {
        if (sockets[x] != -1)
            write(sockets[x], string, strlen(string));
    }

    if (foreground)
        fprintf(stdout, "%s %s", asciitime(&timep), string);

}

void cycle_on_connections(int accept_socket)
{

    char line[BUFLEN];
    fd_set set_fds;
    struct timeval valtime;
    int select_result;
    int len;
    int x;
    char *buffer;
    char *args;
    time_t now, morning, night;
    COMMANDLIST *node;
    char *info, *errorInfo;

    sock = do_accept(accept_socket);

    if (sock) {
        for (x = 0; x < MAXSOCKETS; x++) {
            if (sockets[x] == -1) {
                sockets[x] = sock;
                sprintf(line, "LOGON %d\n", x);
                send_to_all(line);
                break;
            }
        }
    }

    for (x = 0; x < MAXSOCKETS; x++) {

        if (sockets[x] == -1)
            continue;

        sock = sockets[x];

        valtime.tv_sec = SERVER_TIME_OUT_SELECT_SECONDS;
        valtime.tv_usec = SERVER_TIME_OUT_SELECT_USECONDS;

        FD_ZERO(&set_fds);
        FD_SET(sock, &set_fds);

        select_result = select(sock + 1, &set_fds, NULL, NULL, &valtime);

        if (!select_result)
            continue;

        if ((len = read(sock, line, BUFLEN - 1)) <= 0) {
            sockets[x] = -1;
            sprintf(line, "QUIT %d\n", x);
            send_to_all(line);
        } else {
            line[len--] = 0;

            while (len && ((line[len] == '\r') || (line[len] == '\n')))
                line[len--] = 0;

            buffer = strtok((char *) line, " ");
            args = nexttok();

            if (!strcasecmp(buffer, "SEND")) {
                do_send(sock);
            } else if (!strcasecmp(buffer, "STATUS_ALL")) {
                do_status_all(sock);
            } else if (!strcasecmp(buffer, "STATUS_HOUSE")) {
                do_status_house(sock);
            } else if (!strcasecmp(buffer, "STATUS")) {
                do_status(sock);
            } else if (!strcasecmp(buffer, "SHUTDOWN")) {
                running = 0;
            } else if (!strcasecmp(buffer, "TIMERS")) {
                showtimers(sock);
            } else if (!strcasecmp(buffer, "TIME")) {
                now = time(NULL);
                sprintf(line, "TIME %s\n", asciitime(&now));
                write(sock, line, strlen(line));
            } else if (!strcasecmp(buffer, "REHASH")) {
                x10_tcl_init();
            } else if (!strcasecmp(buffer, "QUIT") || !strcasecmp(buffer, "EXIT")) {
                shutdown(sockets[x], 2);
                close(sockets[x]);
            } else if (!strcasecmp(buffer, "PING")) {
                sprintf(line, "PONG\n");
                write(sockets[x], line, strlen(line));
            } else if (!strcasecmp(buffer, "TURN")) {
                do_turn(sockets[x], TOGGLE);
            } else if (!strcasecmp(buffer, "DIM")) {
                do_turn(sockets[x], DIM);
            } else if (!strcasecmp(buffer, "BRIGHT")) {
                do_turn(sockets[x], BRIGHT);
            } else if (!strcasecmp(buffer, "SUNRISE")) {
                now = time(NULL);
                sunshine(latitude, longitude, now, &morning, &night);
                sprintf(line, "SUNRISE %d (%s)\n", (int) morning, asciitime(&morning));
                write(sock, line, strlen(line));
            } else if (!strcasecmp(buffer, "SUNSET")) {
                now = time(NULL);
                sunshine(latitude, longitude, now, &morning, &night);
                sprintf(line, "SUNSET %d (%s)\n", (int) night, asciitime(&night));
                write(sock, line, strlen(line));
            } else if (!strcasecmp(buffer, "UPTIME")) {
                sprintf(line, "UPTIME %s\n", asciitime(&uptime));
                write(sock, line, strlen(line));
            } else if (!strcasecmp(buffer, "EVAL")) {
                if (args != NULL) {
                    if (Tcl_Eval(interp, args) == TCL_ERROR) {
                        sprintf(line, "ERROR (%s) %s\n", args, interp->result);
                        write(sock, line, strlen(line));
                        errorInfo = Tcl_GetVar(interp, "errorInfo", TCL_GLOBAL_ONLY);
                        if (errorInfo != NULL) {
                            info = strtok(errorInfo, "\n");
                            while (info != NULL) {
                                sprintf(line, "DEBUG   %s\n", info);
                                write(sock, line, strlen(line));
                                info = strtok(NULL, "\n");
                            }
                        }
                    } else {
                        sprintf(line, "%s\n", interp->result);
                        write(sock, line, strlen(line));
                    }
                }
            } else {
                if (addedcommands != NULL) {
                    thissocket = sock;
                    node = addedcommands;
                    while (node != NULL) {
                        if (!strcasecmp(buffer, node->command)) {
                            if (args == NULL)
                                sprintf(line, "%s", node->process);
                            else
                                sprintf(line, "%s %s", node->process, args);
                            if (Tcl_Eval(interp, line) == TCL_ERROR) {
                                sprintf(line, "ERROR (%s) %s\n", node->process, interp->result);
                                write(sock, line, strlen(line));
                                errorInfo = Tcl_GetVar(interp, "errorInfo", TCL_GLOBAL_ONLY);
                                if (errorInfo != NULL) {
                                    info = strtok(errorInfo, "\n");
                                    while (info != NULL) {
                                        sprintf(line, "DEBUG   %s\n", info);
                                        write(sock, line, strlen(line));
                                        info = strtok(NULL, "\n");
                                    }
                                }
                            }
                            thissocket = 0;
                            return;
                        } else
                            node = node->child;
                    }
                }
                sprintf(line, "ERROR\n");
                write(sockets[x], line, strlen(line));
                thissocket = 0;
            }
        }
    }

}

int init_server(void)
{
    struct sockaddr_in server;
    int s;

    memset((char *) &server, 0, sizeof(server));

    server.sin_family = AF_INET;
    server.sin_port = htons((unsigned short) port);
    server.sin_addr.s_addr = INADDR_ANY;

    if ((s = socket(AF_INET, SOCK_STREAM, 0)) == -1)
        fatal_error("socket returned is invalid !!\n");

    if ((fcntl(s, F_SETFL, O_NONBLOCK)))
        fatal_error("fcntl O_NONBLOCK failed !!\n");

    if (bind(s, (struct sockaddr *) &server, sizeof(server)) == -1)
        fatal_error("bind failed !!\n");

    if (listen(s, 128) == -1)
        fatal_error("listen failed !!\n");

    return s;
}
