/* x10bot
 * Copyright 2001-2002 Daniel P. Stasinski <mooooooo@scriptkitties.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <time.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <x10_api.h>
#include <tcl.h>
#include <tclDecls.h>
#include "x10bot.h"

extern Tcl_Interp *interp;
extern struct x10_status_t *x10_status;
extern x10module modules[MAXMODULES];
extern event events[MAXEVENTS];
extern char *logfile;
extern int foreground;
extern time_t uptime;
extern double latitude;
extern double longitude;
extern COMMANDLIST *addedcommands;
extern int thissocket;

int tcl_log(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{
    FILE *fp;
    time_t timep;

    BADARGS(2, 2, " text");

    if (logfile != NULL) {
        timep = time(NULL);
        fp = fopen(logfile, "a");
        fprintf(fp, "%s %s", asciitime(&timep), argv[2]);
        fclose(fp);
    }

    if (foreground)
        fprintf(stdout, "%s %s", asciitime(&timep), argv[2]);

    return TCL_OK;
}

int tcl_coordinates(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{
    BADARGS(3, 3, " latitude longitude");
    if (Tcl_GetDouble(irp, argv[1], (double *) &latitude) != TCL_OK) {
        return TCL_ERROR;
    }
    if (Tcl_GetDouble(irp, argv[2], (double *) &longitude) != TCL_OK) {
        return TCL_ERROR;
    }
    return TCL_OK;
}

int tcl_turn(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{

    char script[BUFLEN];
    int command;
    int module_id;
    char house;
    int unit;
    int isgroup;

    BADARGS(3, 4, " module on|off ?housecode?");

    if (strcasecmp(argv[1], "all") == 0) {
        BADARGS(4, 4, " all off housecode");
        module_id = gethouse(argv[3]);
        if (module_id == -1) {
            BADMODULE(argv[3]);
        }
        unit = 1;
        if (strcasecmp(argv[2], "off") == 0)
            command = X10_CMD_ALL_UNITS_OFF;
        else
            return TCL_ERROR;

    } else if (strcasecmp(argv[1], "lights") == 0) {

        BADARGS(4, 4, " lights on|off housecode");
        module_id = gethouse(argv[3]);
        if (module_id == -1) {
            BADMODULE(argv[3]);
        }
        unit = 1;
        if (strcasecmp(argv[2], "on") == 0)
            command = X10_CMD_ALL_LIGHTS_ON;
        else if (strcasecmp(argv[2], "off") == 0)
            command = X10_CMD_ALL_LIGHTS_OFF;
        else
            return TCL_ERROR;

    } else {

        if (strcasecmp(argv[2], "on") == 0)
            command = X10_CMD_ON;
        else if (strcasecmp(argv[2], "off") == 0)
            command = X10_CMD_OFF;
        else
            return TCL_ERROR;

        module_id = getmodule(argv[1]);
        if (module_id == -1) {
            isgroup = FALSE;
            for (module_id = 0; module_id < MAXMODULES; module_id++) {
                if (strcasecmp(argv[1], modules[module_id].group) == 0) {
                    house = ((module_id / 16) + 'A');
                    unit = (module_id % 16) + 1;
                    if (x10_send_cmd(house, unit, command, 0, x10_status)) {
                        sprintf(script, "turn %s %s %s", argv[1], argv[2], argv[3]);
                        Tcl_AppendResult(irp, script, NULL);
                        insert_event(1, SCHED_PROC, 1, 0, 0, -1, script);       /* retry in 1 seconds */
                        isgroup = TRUE;
                    }
                }
            }
            if (!isgroup)
                BADMODULE(argv[1]);
        }

        house = ((module_id / 16) + 'A');
        unit = (module_id % 16) + 1;

    }

    if (x10_send_cmd(house, unit, command, 0, x10_status)) {
        sprintf(script, "turn %s %s %s", argv[1], argv[2], argv[3]);
        Tcl_AppendResult(irp, script, NULL);
        insert_event(1, SCHED_PROC, 1, 0, 0, -1, script);       /* retry in 1 seconds */
        return TCL_OK;
    }

    if (command == X10_CMD_ON)
        camera_group_off(module_id, x10_status);

    return TCL_OK;
}

int tcl_dim(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{

    char script[BUFLEN];
    int module_id;
    char house;
    int unit;
    int value = 0;

    BADARGS(2, 3, " module ?value?");

    module_id = getmodule(argv[1]);
    if (module_id == -1) {
        BADMODULE(argv[1]);
    }

    if (argc > 2) {
        if (Tcl_GetInt(irp, argv[2], &value) != TCL_OK) {
            return TCL_ERROR;
        }
    }

    house = ((module_id / 16) + 'A');
    unit = (module_id % 16) + 1;

    if (x10_send_cmd(house, unit, X10_CMD_DIM, value, x10_status)) {
        sprintf(script, "dim %s %d", argv[1], value);
        Tcl_AppendResult(irp, script, NULL);
        insert_event(1, SCHED_PROC, 1, 0, 0, -1, script);
    }

    return TCL_OK;

}

int tcl_bright(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{

    char script[BUFLEN];
    int module_id;
    char house;
    int unit;
    int value = 0;

    BADARGS(2, 3, " module ?value?");

    module_id = getmodule(argv[1]);
    if (module_id == -1) {
        BADMODULE(argv[1]);
    }

    if (argc > 2) {
        if (Tcl_GetInt(irp, argv[2], &value) != TCL_OK) {
            return TCL_ERROR;
        }
    }

    house = ((module_id / 16) + 'A');
    unit = (module_id % 16) + 1;

    if (x10_send_cmd(house, unit, X10_CMD_BRIGHT, value, x10_status)) {
        sprintf(script, "dim %s %d", argv[1], value);
        Tcl_AppendResult(irp, script, NULL);
        insert_event(1, SCHED_PROC, 1, 0, 0, -1, script);
    }

    return TCL_OK;

}

int tcl_alias(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{

    int module_id;
    int size;

    BADARGS(4, 4, " module aliasname description");

    module_id = getmodule(argv[1]);
    if (module_id == -1) {
        BADMODULE(argv[1]);
    }

    if (modules[module_id].alias != NULL)
        Tcl_Free(modules[module_id].alias);

    size = strlen(argv[2]);

    if (size) {
        modules[module_id].alias = Tcl_Alloc(size + 1);
        strcpy(modules[module_id].alias, argv[2]);
    } else
        modules[module_id].alias = NULL;

    if (modules[module_id].description != NULL)
        Tcl_Free(modules[module_id].description);

    size = strlen(argv[3]);

    if (size) {
        modules[module_id].description = Tcl_Alloc(size + 1);
        strcpy(modules[module_id].description, argv[3]);
    } else
        modules[module_id].description = NULL;

    return TCL_OK;
}

int tcl_group(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{

    int module_id;
    int size;

    BADARGS(3, 3, " module groupname");

    module_id = getmodule(argv[1]);
    if (module_id == -1) {
        BADMODULE(argv[1]);
    }

    if (modules[module_id].group != NULL)
        Tcl_Free(modules[module_id].group);

    size = strlen(argv[2]);

    if (size) {
        modules[module_id].group = Tcl_Alloc(size + 1);
        strcpy(modules[module_id].group, argv[2]);
    } else
        modules[module_id].group = NULL;

    return TCL_OK;
}

int tcl_bind(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{

    int module_id;
    int size;

    BADARGS(3, 3, " module proc");

    module_id = getmodule(argv[1]);
    if (module_id == -1) {
        BADMODULE(argv[1]);
    }

    if (modules[module_id].process != NULL)
        Tcl_Free(modules[module_id].process);

    size = strlen(argv[2]);

    if (size) {
        modules[module_id].process = Tcl_Alloc(size + 1);
        strcpy(modules[module_id].process, argv[2]);
    } else
        modules[module_id].process = NULL;

    Tcl_AppendResult(irp, argv[1], NULL);

    return TCL_OK;
}

int tcl_getstatus(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{

    int module_id;
    char house;
    int unit;
    struct x10_object_t module;
    char reply[64];

    BADARGS(2, 2, " module");

    module_id = getmodule(argv[1]);
    if (module_id == -1) {
        BADMODULE(argv[1]);
    }

    house = module_id / 16;
    unit = module_id % 16;

    module = x10_status->objects[(int) house][unit];

    sprintf(reply, "%s %d", status2string(module.status), (int) module.timestamp);

    Tcl_AppendResult(irp, reply, NULL);

    return TCL_OK;
}

int tcl_setstatus(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{

    int module_id;
    char house;
    int unit;
    struct x10_object_t module;
    char reply[64];

    BADARGS(3, 3, " module status");

    module_id = getmodule(argv[1]);
    if (module_id == -1) {
        BADMODULE(argv[1]);
    }

    house = (module_id / 16);
    unit = module_id % 16;

    module = x10_status->objects[(int) house][unit];

    module.timestamp = time(NULL);
    module.status = string2status(argv[2]);

    sprintf(reply, "%s %d", status2string(module.status), (int) module.timestamp);

    Tcl_AppendResult(irp, reply, NULL);

    return TCL_OK;

}

int tcl_getmodule(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{

    int module_id;
    char house;
    int unit;
    char reply[32];

    BADARGS(2, 2, " module");

    module_id = getmodule(argv[1]);
    if (module_id == -1) {
        BADMODULE(argv[1]);
    }

    house = ((module_id / 16) + 'A');
    unit = (module_id % 16) + 1;

    sprintf(reply, "%c%d", house, unit);

    Tcl_AppendResult(irp, reply, NULL);

    return TCL_OK;
}

int tcl_getalias(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{

    int module_id;

    BADARGS(2, 2, " module");

    module_id = getmodule(argv[1]);
    if (module_id == -1) {
        BADMODULE(argv[1]);
    }

    if (modules[module_id].alias != NULL)
        Tcl_AppendResult(irp, modules[module_id].alias, NULL);
    else
        Tcl_AppendResult(irp, argv[1], NULL);

    return TCL_OK;
}

int tcl_getdescription(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{

    int module_id;

    BADARGS(2, 2, " module");

    module_id = getmodule(argv[1]);
    if (module_id == -1) {
        BADMODULE(argv[1]);
    }

    if (modules[module_id].description != NULL)
        Tcl_AppendResult(irp, modules[module_id].description, NULL);

    return TCL_OK;
}


int tcl_getbind(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{

    int module_id;

    BADARGS(2, 2, " module");

    module_id = getmodule(argv[1]);
    if (module_id == -1) {
        BADMODULE(argv[1]);
    }

    if (modules[module_id].process != NULL)
        Tcl_AppendResult(irp, modules[module_id].process, NULL);

    return TCL_OK;
}

int tcl_getgroup(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{

    int module_id;

    BADARGS(2, 2, " module");

    module_id = getmodule(argv[1]);
    if (module_id == -1) {
        BADMODULE(argv[1]);
    }

    if (modules[module_id].group != NULL)
        Tcl_AppendResult(irp, modules[module_id].group, NULL);

    return TCL_OK;
}

int tcl_schedule_once(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{
    char *script;
    int timestamp;
    int first;
    int handle;
    int thread = 0, quiet = 0;
    char reply[32];

    BADARGS(3, 5, " ?-thread? ?-quiet? unixtime {script | procname ?args?}");

    first = 1;

    while (argc--) {
        if (strcasecmp(argv[first], "-thread") == 0) {
            thread = 1;
        } else if (strcasecmp(argv[first], "-quiet") == 0) {
            quiet = 1;
        } else {
            break;
        }
        first++;
    }

    if (Tcl_GetInt(irp, argv[first++], &timestamp) != TCL_OK) {
        return TCL_ERROR;
    }

    script = argv[first];

    if ((handle = insert_event(timestamp, SCHED_ONCE, 1, thread, quiet, -1, script)) == -1) {
        Tcl_AppendResult(irp, "Event queue full", NULL);
        return TCL_ERROR;
    }

    sprintf(reply, "%d", handle);
    Tcl_AppendResult(irp, reply, NULL);
    return TCL_OK;
}

int tcl_schedule_proc(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{
    char *script;
    int seconds;
    int first;
    int handle;
    int once = 0, thread = 0, quiet = 0, repeat = -1;
    char reply[32];

    BADARGS(3, 8, " ?-thread? ?-once? ?-quiet? ?-repeat count? interval {script | procname ?args?}");

    first = 1;

    while (argc--) {
        if (strcasecmp(argv[first], "-thread") == 0) {
            thread = 1;
        } else if (strcasecmp(argv[first], "-once") == 0) {
            once = 1;
        } else if (strcasecmp(argv[first], "-quiet") == 0) {
            quiet = 1;
        } else if (strcasecmp(argv[first], "-repeat") == 0) {
            if (Tcl_GetInt(interp, argv[first + 1], &repeat) != TCL_OK) {
                Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " ?-thread? ?-once? ?-quiet? ?-repeat count? interval {script | procname ?args?}", NULL);
                return TCL_ERROR;
            }
            first++;
            argc--;
        } else {
            break;
        }
        first++;
    }

    if (Tcl_GetInt(irp, argv[first++], &seconds) != TCL_OK) {
        return TCL_ERROR;
    }

    script = argv[first];

    if ((handle = insert_event(seconds, SCHED_PROC, once, thread, quiet, repeat, script)) == -1) {
        Tcl_AppendResult(irp, "Event queue full", NULL);
        return TCL_ERROR;
    }

    sprintf(reply, "%d", handle);
    Tcl_AppendResult(irp, reply, NULL);
    return TCL_OK;

}

int tcl_schedule_days(ClientData cd, Tcl_Interp * irp, int argc, char *argv[], int mode)
{
    char *script;
    int first;
    int handle;
    int once = 0, thread = 0, quiet = 0, repeat = -1;
    char reply[32];
    int hour, minute;
    int seconds;

    BADARGS(4, 9, " ?-thread? ?-once? ?-quiet? ?-repeat count? hour minute {script | procname ?args?}");

    first = 1;

    while (argc--) {
        if (strcasecmp(argv[first], "-thread") == 0) {
            thread = 1;
        } else if (strcasecmp(argv[first], "-once") == 0) {
            once = 1;
        } else if (strcasecmp(argv[first], "-quiet") == 0) {
            quiet = 1;
        } else if (strcasecmp(argv[first], "-repeat") == 0) {
            if (Tcl_GetInt(interp, argv[first + 1], &repeat) != TCL_OK) {
                Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " ?-thread? ?-once? ?-quiet? ?-repeat count? interval {script | procname ?args?}", NULL);
                return TCL_ERROR;
            }
            first++;
            argc--;
        } else {
            break;
        }
        first++;
    }

    if (Tcl_GetInt(irp, argv[first++], &hour) != TCL_OK) {
        return TCL_ERROR;
    }

    if (hour < 0 || hour > 23) {
        Tcl_AppendResult(irp, "invalid hour \"", argv[first - 1], "\": should be >= 0 and <= 23", NULL);
        return TCL_ERROR;
    }

    if (Tcl_GetInt(irp, argv[first++], &minute) != TCL_OK) {
        return TCL_ERROR;
    }

    if (minute < 0 || minute > 59) {
        Tcl_AppendResult(irp, "invalid minute \"", argv[first - 1], "\": should be >= 0 and <= 59", NULL);
        return TCL_ERROR;
    }

    seconds = (hour * 3600) + (minute * 60);

    script = argv[first];

    if ((handle = insert_event(seconds, mode, once, thread, quiet, repeat, script)) == -1) {
        Tcl_AppendResult(irp, "Event queue full", NULL);
        return TCL_ERROR;
    }

    sprintf(reply, "%d", handle);
    Tcl_AppendResult(irp, reply, NULL);
    return TCL_OK;

}

int tcl_schedule_daily(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{
    return (tcl_schedule_days(cd, irp, argc, argv, SCHED_DAILY));
}

int tcl_schedule_weekday(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{
    return (tcl_schedule_days(cd, irp, argc, argv, SCHED_WEEKDAY));
}

int tcl_schedule_weekend(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{
    return (tcl_schedule_days(cd, irp, argc, argv, SCHED_WEEKEND));
}

int tcl_schedule_weekly(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{
    char *script;
    int first;
    int handle;
    int once = 0, thread = 0, quiet = 0, repeat = -1;
    char reply[32];
    int hour, minute, day;
    int seconds;

    BADARGS(5, 10, " ?-thread? ?-once? ?-quiet? ?-repeat count? day hour minute {script | procname ?args?}");

    first = 1;

    while (argc--) {
        if (strcasecmp(argv[first], "-thread") == 0) {
            thread = 1;
        } else if (strcasecmp(argv[first], "-once") == 0) {
            once = 1;
        } else if (strcasecmp(argv[first], "-quiet") == 0) {
            quiet = 1;
        } else if (strcasecmp(argv[first], "-repeat") == 0) {
            if (Tcl_GetInt(interp, argv[first + 1], &repeat) != TCL_OK) {
                Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " ?-thread? ?-once? ?-quiet? ?-repeat count? interval {script | procname ?args?}", NULL);
                return TCL_ERROR;
            }
            first++;
            argc--;
        } else {
            break;
        }
        first++;
    }

    if (Tcl_GetInt(irp, argv[first++], &day) != TCL_OK) {
        return TCL_ERROR;
    }
    if (day < 0 || day > 6) {
        Tcl_AppendResult(irp, "invalid day \"", argv[first - 1], "\": should be >= 0 and <= 6", NULL);
        return TCL_ERROR;
    }

    if (Tcl_GetInt(irp, argv[first++], &hour) != TCL_OK) {
        return TCL_ERROR;
    }
    if (hour < 0 || hour > 23) {
        Tcl_AppendResult(irp, "invalid hour \"", argv[first - 1], "\": should be >= 0 and <= 23", NULL);
        return TCL_ERROR;
    }
    if (Tcl_GetInt(irp, argv[first++], &minute) != TCL_OK) {
        return TCL_ERROR;
    }
    if (minute < 0 || minute > 59) {
        Tcl_AppendResult(irp, "invalid minute \"", argv[first - 1], "\": should be >= 0 and <= 59", NULL);
        return TCL_ERROR;
    }

    seconds = (((day * 24) + hour) * 3600) + (minute * 60);

    script = argv[first];

    if ((handle = insert_event(seconds, SCHED_WEEKLY, once, thread, quiet, repeat, script)) == -1) {
        Tcl_AppendResult(irp, "Event queue full", NULL);
        return TCL_ERROR;
    }

    sprintf(reply, "%d", handle);
    Tcl_AppendResult(irp, reply, NULL);
    return TCL_OK;

}


int tcl_schedule_exists(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{
    int x;
    int handle;

    BADARGS(2, 2, " schedid");

    if (Tcl_GetInt(irp, argv[1], &handle) != TCL_OK) {
        return TCL_ERROR;
    }

    for (x = 0; x < MAXEVENTS; x++) {
        if (events[x].mode != SCHED_NEVER) {
            if (events[x].handle == handle) {
                Tcl_AppendResult(irp, "1", NULL);
                return TCL_OK;
            }
        }
    }

    Tcl_AppendResult(irp, "0", NULL);

    return TCL_OK;

}


int tcl_unschedule_proc(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{
    int x;
    int handle;

    BADARGS(2, 2, " schedid");

    if (Tcl_GetInt(irp, argv[1], &handle) != TCL_OK) {
        return TCL_ERROR;
    }

    for (x = 0; x < MAXEVENTS; x++) {
        if (events[x].mode != SCHED_NEVER) {
            if (events[x].handle == handle) {
                events[x].mode = SCHED_NEVER;
                if (events[x].script != NULL)
                    Tcl_Free(events[x].script);
                events[x].script = NULL;
                break;
            }
        }
    }

    return TCL_OK;

}

int tcl_sunrise(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{
    char reply[32];
    time_t now, morning, night;
    BADARGS(1, 2, " unixtime");

    if (argc == 2) {
        if (Tcl_GetInt(irp, argv[1], (int *) &now) != TCL_OK) {
            return TCL_ERROR;
        }
    } else {
        now = time(NULL);
    }

    sunshine(latitude, longitude, now, &morning, &night);
    sprintf(reply, "%d", (int) morning);
    Tcl_AppendResult(irp, reply, NULL);
    return TCL_OK;
}

int tcl_sunset(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{
    char reply[32];
    time_t now, morning, night;
    BADARGS(1, 2, " unixtime");

    if (argc == 2) {
        if (Tcl_GetInt(irp, argv[1], (int *) &now) != TCL_OK) {
            return TCL_ERROR;
        }
    } else {
        now = time(NULL);
    }

    sunshine(latitude, longitude, now, &morning, &night);

    sprintf(reply, "%d", (int) night);
    Tcl_AppendResult(irp, reply, NULL);
    return TCL_OK;
}

int tcl_daytime(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{
    time_t now, morning, night;
    BADARGS(1, 2, " unixtime");

    if (argc == 2) {
        if (Tcl_GetInt(irp, argv[1], (int *) &now) != TCL_OK) {
            return TCL_ERROR;
        }
    } else {
        now = time(NULL);
    }

    sunshine(latitude, longitude, now, &morning, &night);

    if ((now >= morning) && (now <= night))
        Tcl_AppendResult(irp, "1", NULL);
    else
        Tcl_AppendResult(irp, "0", NULL);

    return TCL_OK;
}

int tcl_transmit(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{
    BADARGS(2, 2, " text");
    send_to_all(argv[1]);
    return TCL_OK;
}

int tcl_echo(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{
    BADARGS(2, 2, " text");
    if (thissocket)
        write(thissocket, argv[1], strlen(argv[1]));
    return TCL_OK;
}

int tcl_toseconds(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{
    int section = 0;
    int seconds = 0;
    int value;
    char reply[32];

    BADARGS(2, 5, " ?days? ?hours? ?minutes? seconds");

    while (--argc) {

        if (Tcl_GetInt(irp, argv[argc], (int *) &value) != TCL_OK) {
            return TCL_ERROR;
        }

        switch (section++) {
        case 0:
            seconds = value;
            break;
        case 1:
            seconds += (value * 60);
            break;
        case 2:
            seconds += (value * 3600);
            break;
        case 3:
            seconds += (value * 86400);
            break;
        }
    }

    sprintf(reply, "%d", (int) seconds);
    Tcl_AppendResult(irp, reply, NULL);

    return TCL_OK;

}

int tcl_uptime(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{
    char reply[32];
    sprintf(reply, "%d", (int) uptime);
    Tcl_AppendResult(irp, reply, NULL);
    return TCL_OK;
}

int tcl_powerfail(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{
    char reply[32];
    BADARGS(1, 2, " ?-set? ?-reset?");

    if (argc == 2) {
        if (strcasecmp(argv[1], "-set") == 0) {
            x10_status->power_fail = 1;
        } else if (strcasecmp(argv[1], "-reset") == 0) {
            x10_status->power_fail = 0;
        }
    }

    sprintf(reply, "%i", x10_status->power_fail);
    Tcl_AppendResult(irp, reply, NULL);
    return TCL_OK;
}

int tcl_addcommand(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{
    int size;
    COMMANDLIST *node, *newnode;

    BADARGS(3, 3, " command proc");

    newnode = (COMMANDLIST *) Tcl_Alloc(sizeof(COMMANDLIST));
    newnode->child = NULL;

    if (addedcommands == NULL) {
        newnode->parent = NULL;
        addedcommands = newnode;
    } else {
        node = addedcommands;
        while (node->child != NULL) {
            node = node->child;
        }
        node->child = newnode;
        newnode->parent = node;
    }

    size = strlen(argv[1]);
    newnode->command = Tcl_Alloc(size + 1);
    strcpy(newnode->command, argv[1]);

    size = strlen(argv[2]);
    newnode->process = Tcl_Alloc(size + 1);
    strcpy(newnode->process, argv[2]);

    return TCL_OK;
}

int tcl_getschedule(ClientData cd, Tcl_Interp * irp, int argc, char *argv[])
{

    int x, handle;
    char reply[BUFLEN];

    BADARGS(1, 2, " schedid");

    reply[0] = '\0';

    if (argc == 2) {

        if (Tcl_GetInt(irp, argv[1], &handle) != TCL_OK) {
            return TCL_ERROR;
        }

        for (x = 0; x < MAXEVENTS; x++) {
            if (events[x].mode != SCHED_NEVER) {
                if (events[x].handle == handle) {
                    sprintf(reply, "%d %d {%s}", events[x].handle, (int) events[x].timestamp, events[x].script);
                    Tcl_AppendResult(irp, reply, NULL);
                    return TCL_OK;
                }
            }
        }
    } else {
        for (x = 0; x < MAXEVENTS; x++) {
            if ((events[x].mode == SCHED_NEVER) || (events[x].script == NULL))
                continue;
            sprintf(reply, "{%d %d {%s}} ", events[x].handle, (int) events[x].timestamp, events[x].script);
            Tcl_AppendResult(irp, reply, NULL);
        }
    }

    return TCL_OK;
}

void x10_tcl_init(void)
{
    int x;
    char reply[BUFLEN];
    char *info, *errorInfo;
    COMMANDLIST *node;

    for (x = 0; x < MAXMODULES; x++) {
        if (modules[x].alias != NULL)
            Tcl_Free(modules[x].alias);
        if (modules[x].description != NULL)
            Tcl_Free(modules[x].description);
        if (modules[x].process != NULL)
            Tcl_Free(modules[x].process);
        if (modules[x].group != NULL)
            Tcl_Free(modules[x].group);
        modules[x].alias = NULL;
        modules[x].description = NULL;
        modules[x].process = NULL;
        modules[x].group = NULL;
    }

    modules[256].alias = Tcl_Alloc(4);
    strcpy(modules[256].alias, "all");
    modules[257].alias = Tcl_Alloc(7);
    strcpy(modules[257].alias, "lights");

    for (x = 0; x < MAXEVENTS; x++) {
        events[x].mode = SCHED_NEVER;
        if (events[x].script != NULL)
            Tcl_Free(events[x].script);
        events[x].script = NULL;
    }

    if (addedcommands != NULL) {
        node = addedcommands;
        while (node != NULL) {
            if (node->command != NULL)
                Tcl_Free(node->command);
            if (node->process != NULL)
                Tcl_Free(node->process);
            if (node->parent != NULL)
                Tcl_Free((char *) node->parent);
            if (node->child != NULL)
                node = node->child;
            else {
                Tcl_Free((char *) node);
                node = NULL;
            }
        }
        addedcommands = NULL;
    }

    Tcl_CreateCommand(interp, "turn", tcl_turn, NULL, NULL);
    Tcl_CreateCommand(interp, "bright", tcl_bright, NULL, NULL);
    Tcl_CreateCommand(interp, "dim", tcl_dim, NULL, NULL);
    Tcl_CreateCommand(interp, "getstatus", tcl_getstatus, NULL, NULL);
    Tcl_CreateCommand(interp, "setstatus", tcl_setstatus, NULL, NULL);
    Tcl_CreateCommand(interp, "alias", tcl_alias, NULL, NULL);
    Tcl_CreateCommand(interp, "group", tcl_group, NULL, NULL);
    Tcl_CreateCommand(interp, "bind", tcl_bind, NULL, NULL);
    Tcl_CreateCommand(interp, "getmodule", tcl_getmodule, NULL, NULL);
    Tcl_CreateCommand(interp, "getalias", tcl_getalias, NULL, NULL);
    Tcl_CreateCommand(interp, "getdescription", tcl_getdescription, NULL, NULL);
    Tcl_CreateCommand(interp, "getbind", tcl_getbind, NULL, NULL);
    Tcl_CreateCommand(interp, "getgroup", tcl_getgroup, NULL, NULL);
    Tcl_CreateCommand(interp, "schedule_proc", tcl_schedule_proc, NULL, NULL);
    Tcl_CreateCommand(interp, "schedule_once", tcl_schedule_once, NULL, NULL);
    Tcl_CreateCommand(interp, "unschedule_proc", tcl_unschedule_proc, NULL, NULL);
    Tcl_CreateCommand(interp, "schedule_daily", tcl_schedule_daily, NULL, NULL);
    Tcl_CreateCommand(interp, "schedule_weekly", tcl_schedule_weekly, NULL, NULL);
    Tcl_CreateCommand(interp, "schedule_weekday", tcl_schedule_weekday, NULL, NULL);
    Tcl_CreateCommand(interp, "schedule_weekend", tcl_schedule_weekend, NULL, NULL);
    Tcl_CreateCommand(interp, "schedule_exists", tcl_schedule_exists, NULL, NULL);
    Tcl_CreateCommand(interp, "sunset", tcl_sunset, NULL, NULL);
    Tcl_CreateCommand(interp, "sunrise", tcl_sunrise, NULL, NULL);
    Tcl_CreateCommand(interp, "daytime", tcl_daytime, NULL, NULL);
    Tcl_CreateCommand(interp, "transmit", tcl_transmit, NULL, NULL);
    Tcl_CreateCommand(interp, "coordinates", tcl_coordinates, NULL, NULL);
    Tcl_CreateCommand(interp, "log", tcl_log, NULL, NULL);
    Tcl_CreateCommand(interp, "toseconds", tcl_toseconds, NULL, NULL);
    Tcl_CreateCommand(interp, "powerfail", tcl_powerfail, NULL, NULL);
    Tcl_CreateCommand(interp, "uptime", tcl_uptime, NULL, NULL);
    Tcl_CreateCommand(interp, "addcommand", tcl_addcommand, NULL, NULL);
    Tcl_CreateCommand(interp, "echo", tcl_echo, NULL, NULL);
    Tcl_CreateCommand(interp, "getschedule", tcl_getschedule, NULL, NULL);

    send_to_all("REHASH\n");

    if (Tcl_EvalFile(interp, "/etc/x10bot.conf") == TCL_ERROR) {
        sprintf(reply, "ERROR (/etc/x10bot.conf) %s\n", interp->result);
        send_to_all(reply);
        errorInfo = Tcl_GetVar(interp, "errorInfo", TCL_GLOBAL_ONLY);
        if (errorInfo != NULL) {
            info = strtok(errorInfo, "\n");
            while (info != NULL) {
                sprintf(reply, "DEBUG   %s\n", info);
                send_to_all(reply);
                info = strtok(NULL, "\n");
            }
        }
    }
}
